#!/usr/bin/env python
# $Id: am.py 1.2 05/04/05 13:24:26-05:00 bradshaw@skywalker.mcs.anl.gov $

from elementtree.ElementTree import Element, SubElement, parse

from sss.restriction import Data, DataSet
from sss.server import Server
from random import random
from sys import argv

class AmountError(Exception):
    pass

class Project(Data):
    def AddUser(self, user):
        if not [x for x in self.element.findall('user') if x.attrib['name'] == user]:
            self.element.append(Element('user', name=user))

    def DelUser(self, user):
        u = [x for x in self.element.findall('user') if x.attrib['name'] == user]
        if len(u) == 1:
            self.element.remove(u[0])

    def AddLien(self, amount):
        amount = float(amount)
        nodeseconds = float(self.element.get('allocation',default='-1'))#if there is no allocation, no limit 
        amount = float(amount)
        if (amount <= nodeseconds):
            nodeseconds -= amount
        elif nodeseconds == -1:
            q=1#unlimited seconds, just needed indented block
        else:
            raise AmountError
            return

        id = random()
        u = [x for x in self.element.findall('lien') if x.attrib['id'] == id]
        while len(u) != 0:
            id = str(random())
            u = [x for x in self.element.findall('lien') if x.attrib['id'] == id]

        self.element.set('allocation',str(nodeseconds))    
        self.element.append(Element('lien',id=str(id),amount=str(amount)))
        return id

    def DelLien(self, id):
        u = [x for x in self.element.findall('lien') if x.attrib['id'] == id]
        nodeseconds = float(self.element.get('allocation')) + float(u[0].get('amount'))
        self.element.set('allocation',str(nodeseconds))
        self.element.remove(u[0])
        
    def CommitLien(self, id, actual):
        u = [x for x in self.element.findall('lien') if x.attrib['id'] == id]
        d = ( float(u[0].get('amount')) -float(actual))
        nodeseconds = float(self.element.get('allocation'))
        if (nodeseconds > -1):
            nodeseconds = float( self.element.get('allocation'))+d
        else:
            nodeseconds = -1
        self.element.set('allocation',str(nodeseconds))
        self.element.remove(u[0])

class allocmgr(Server):
    __component__ = 'allocation-manager'
    __implementation__ = 'chiba-am'
    __statefields__ = ['projects']
    __dispatch__ = {'add-project':"projects.Add", 'get-project':'projects.Get', 'del-project':'projects.Del', 'add-user':'XAddUser','del-user':'XDelUser','add-lien':'XAddLien','del-lien':'XDelLien','commit-lien':'XCommit'}
    __validate__ = 0

    def __setup__(self):
        self.projects=DataSet("projects","project",Project,False,False)
        
    def XAddUser(self, xml, (peer,port)):
        return self.projects.Get(xml, (peer,port), self.AddUser)

    def AddUser(self, proj, args):
        proj.AddUser(args['user'])

    def XDelUser(self, xml, (peer,port)):
        return self.projects.Get(xml, (peer,port), self.DelUser)

    def DelUser(self, proj, args):
        proj.DelUser(args['user'])

    def XAddLien(self, xml, (peer,port)):
        result = Element('projects')
        for s in xml.findall('project'):
            for item in [x for x in self.projects.data if x.Match(s)]:
                id =  self.AddLien(item, xml.attrib)
                result.append(item.Fetch(s))
        if id:result.append(Element('lien',id=str(id)))
        return result

    def AddLien(self, proj, args):
        id = proj.AddLien(args['amount'])
        if id:
            return id
        else:
            return
        
    def XDelLien(self,xml, (peer,port)):
        return self.projects.Get(xml, (peer,port), self.DelLien)

    def DelLien(self, proj, args):
        proj.DelLien(args['id'])

    def XCommit(self, xml, (peer,port)):
        return self.projects.Get(xml, (peer,port), self.CommitLien)

    def CommitLien(self, proj, args):
        proj.CommitLien(args['id'],args['actual'])

if __name__ == '__main__':
    from getopt import getopt, GetoptError
    from sys import argv,exit
    try:
        (opts,args)=getopt(argv[1:],'d',['daemon='])
    except GetoptError,msg:
        print "%s\nUsage:\nam.py [-d] [--daemon <pidfile>]"%(msg)
        exit(1)
    daemon=[x[1] for x in opts if x[0] == '--daemon']
    debug=len([x for x in opts if x[0] == '-d'])
    if daemon:
        from sss.daemonize import daemonize
        daemonize(daemon[0])
    a = allocmgr(debug=debug)
    a.ServeForever()
